import copy
import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar, Literal, Optional

from pydantic import BaseModel, ConfigDict
from pydantic_core import PydanticUndefined

from .constants import ORDERED_SETTINGS, PYPROJECT_TOML_MAX_DEPTH
from .entity import (
    EnvConfigEntity as env,
    FileConfigEntity as conf,
    ModelConfigEntity as model,
)
from .errors import ArFiSettingsError
from .handlers import ArFiBaseHandler, ArFiHandler
from .schemes import (
    CONF_INCLUDE_EXLUDE_PARAMS,
    ENV_INCLUDE_EXLUDE_PARAMS,
    INCLUDE_EXLUDE_PARAMS,
    EnvConfigSchema,
    FileConfigSchema,
    SettingsConfigSchema,
    SettingsParamsSchema,
)
from .types import (
    DEFAULT_PATH_SENTINEL,
    LIST_STR_SENTINEL,
    STR_SENTINEL,
    EnvConfigDict,
    FileConfigDict,
    GlobalConfigDict,
    MultiPathType,
    PathType,
)
from .utils import is_settings

if TYPE_CHECKING:
    from .main import ArFiSettings

init_settings = None


class InstanceConfig(BaseModel):
    root_dir: PathType | None = None
    base_dir: PathType | None = None
    instance: Optional[BaseModel] = None

    read_config: bool = True
    read_config_force: bool | None = None
    read_pyproject_toml: bool = True
    read_pyproject_toml_force: bool | None = None
    pyproject_toml_path: PathType | None = None
    mode_dir: PathType | None = DEFAULT_PATH_SENTINEL
    parent_mode_dir: PathType | None = DEFAULT_PATH_SENTINEL
    mode_dir_path: PathType = DEFAULT_PATH_SENTINEL
    computed_mode_dir: str = ""
    source_mode_dir: str = ""
    nested_mode_dir: PathType | None = DEFAULT_PATH_SENTINEL
    mode_dir_attr: str | None = None
    mode_dir_inherit_nested: bool = True
    mode_dir_inherit_parent: bool = True
    file_config_inherit_parent: bool = True
    env_config_inherit_parent: bool = True

    file_config: FileConfigDict = FileConfigDict()
    env_config: EnvConfigDict = EnvConfigDict()
    global_config: GlobalConfigDict = GlobalConfigDict()
    class_file_config: FileConfigSchema = FileConfigSchema()
    init_file_config: FileConfigSchema = FileConfigSchema()
    parent_file_config: FileConfigSchema = FileConfigSchema()
    computed_file_config: FileConfigDict = {}
    conf_path: list[Path] = []

    class_env_config: EnvConfigSchema = EnvConfigSchema()
    init_env_config: EnvConfigSchema = EnvConfigSchema()
    parent_env_config: EnvConfigSchema = EnvConfigSchema()
    computed_env_config: EnvConfigDict = {}
    env_path: list[Path] = []

    handler_class: type[ArFiBaseHandler] = ArFiHandler
    handler: str = "default_main_handler"
    handler_inherit_parent: bool = True
    ordered_settings: list[str] = ORDERED_SETTINGS
    ordered_settings_inherit_parent: bool = True

    settings_config: Optional[SettingsConfigSchema] = None
    class_settings_config: Optional[SettingsConfigSchema] = None
    settings_init_params: SettingsParamsSchema = SettingsParamsSchema()
    values_by_default: dict[str, Any] = {}
    init_kwargs: dict[str, Any] = {}
    handler_tree: list[list[str]] = []
    handler_parent_mode_dir: list[str | Path] = []

    modified_class_vars: set[str] = set()
    inherited_params: list[str] = []
    search_base_dir: bool = True

    _class_vars: ClassVar[set[str]] = {
        "instance",
        "class_file_config",
        "class_env_config",
        "class_settings_config",
        "handler_class",
        "arfi_dev_debug",
        "modified_class_vars",
        "init_kwargs",
        "file_config",
        "env_config",
        "global_config",
        "read_pyproject_toml_force",
    }
    model_config = ConfigDict(
        extra="ignore",
        validate_assignment=True,
        validate_default=False,
        validate_return=False,
    )

    # TODO: append read arfi_dev_debug from ./config/arfi_dev_debug.toml
    arfi_dev_debug: bool = False
    arfi_debug: bool = False

    def load(
        self,
        instance: "ArFiSettings",
        _read_config: bool = None,
        _read_config_force: bool = None,
        _read_pyproject_toml: bool | None = None,
        _pyproject_toml_depth: int | None = None,
        _pyproject_toml_max_depth: int = PYPROJECT_TOML_MAX_DEPTH,
        _mode_dir: PathType | None = DEFAULT_PATH_SENTINEL,
        _mode_dir_inherit_nested: bool = None,
        _mode_dir_inherit_parent: bool = None,
        _file_config_inherit_parent: bool = None,
        _env_config_inherit_parent: bool = None,
        _conf_file: MultiPathType | None = DEFAULT_PATH_SENTINEL,
        _conf_dir: MultiPathType | None = DEFAULT_PATH_SENTINEL,
        _conf_ext: str | list[str, ...] = [],
        _conf_file_encoding: str | None = STR_SENTINEL,
        _conf_case_sensitive: bool = None,
        _conf_ignore_missing: bool = None,
        _conf_custom_ext_handler: str | dict[str, str] | None = STR_SENTINEL,
        _conf_include_inherit_parent: list[Literal[*CONF_INCLUDE_EXLUDE_PARAMS], ...] = None,
        _conf_exclude_inherit_parent: list[Literal[*CONF_INCLUDE_EXLUDE_PARAMS], ...] = None,
        _env_file: MultiPathType | None = DEFAULT_PATH_SENTINEL,
        _env_prefix: str = None,
        _env_prefix_as_mode_dir: bool = None,
        _env_prefix_as_nested_mode_dir: bool = None,
        _env_prefix_as_source_mode_dir: bool = None,
        _env_file_encoding: str | None = STR_SENTINEL,
        _env_case_sensitive: bool = None,
        _env_nested_delimiter: str = None,
        _env_ignore_missing: bool = None,
        _env_include_inherit_parent: list[Literal[*ENV_INCLUDE_EXLUDE_PARAMS], ...] = None,
        _env_exclude_inherit_parent: list[Literal[*ENV_INCLUDE_EXLUDE_PARAMS], ...] = None,
        _case_sensitive: bool = None,
        _ignore_missing: bool = None,
        _encoding: str | None = STR_SENTINEL,
        _secrets_dir: PathType | None = DEFAULT_PATH_SENTINEL,
        _include_inherit_parent: list[Literal[*INCLUDE_EXLUDE_PARAMS], ...] = None,
        _exclude_inherit_parent: list[Literal[*INCLUDE_EXLUDE_PARAMS], ...] = None,
        _handler: str = STR_SENTINEL,
        _handler_inherit_parent: bool = None,
        _ordered_settings: list[str, ...] = LIST_STR_SENTINEL,
        _ordered_settings_inherit_parent: bool = None,
        _cli: bool = None,
        **kwargs,
    ) -> None:
        """Load settings configuration and params."""

        global init_settings
        # Don't move to global scope !!!
        from .init_config import init_settings

        values = kwargs.get("values", {})
        self._setup_instance_and_default_value(instance, **values)

        self._setup_params_from_handler(
            values,
            _mode_dir_inherit_parent,
            _file_config_inherit_parent,
            _env_config_inherit_parent,
            _ordered_settings_inherit_parent,
            _handler_inherit_parent,
        )
        if self.mode_dir == DEFAULT_PATH_SENTINEL:
            if self.mode_dir_attr and self.mode_dir_attr != DEFAULT_PATH_SENTINEL:
                self.mode_dir = self.mode_dir_attr
        self.source_mode_dir = self.instance_source_mode_dir
        _kwargs: dict = locals()
        _kwargs.pop("self")
        self.settings_init_params.update_exclude_default(**_kwargs)
        if not self.init_kwargs:
            self.init_kwargs = self.settings_init_params.model_dump(exclude_defaults=True)
        _kwargs = self.settings_init_params.get_param_dict()
        _read_config = _kwargs.get("_read_config", _read_config)
        _read_config_force = _kwargs.get("_read_config_force", _read_config_force)
        self._setup_read_config(
            _read_config=_read_config,
            _read_config_force=_read_config_force,
        )
        handler = _kwargs.pop("_handler")
        if handler:
            self.handler = handler
        if self.values_by_default:
            self.values_by_default.update(values)
            values = self.values_by_default

        ################ PYPROJECT ################
        if _read_pyproject_toml is not None:
            self.read_pyproject_toml = _read_pyproject_toml

        if self._need_read_pyproject(
            _read_config=_read_config,
            _read_config_force=_read_config_force,
        ):
            init_settings.read_pyproject(
                read_pyproject_toml=self.read_pyproject_toml,
                pyproject_toml_depth=_pyproject_toml_depth,
                pyproject_toml_max_depth=_pyproject_toml_max_depth,
                class_name=instance.__class__.__name__,
                search_base_dir=self.search_base_dir,
            )
        if self.pyproject_toml_path != init_settings.pyproject_toml_path:
            if self.pyproject_toml_path is not None:
                warnings.warn_explicit(
                    f"\033[33m\n"
                    f"Path to pyproject.toml has been changed !!!\n"
                    f"for instance {instance.__class__.__name__}()\n"
                    f"inside class {init_settings.main_config_class}\n"
                    f"    previous path:\n"
                    f"{self.pyproject_toml_path.as_posix()}\n"
                    f"    current path:\n"
                    f"{init_settings.pyproject_toml_path.as_posix()}\n"
                    f"Call once\n"
                    f"  from arfi_settings.init_config import init_settings\n"
                    f"  init_settings.read_pyproject(read_once=True)\n"
                    f"before import any instance or subclass `ArFiSettings` for fix it."
                    f"\033[0m",
                    category=Warning,
                    filename=init_settings.main_config_file.as_posix(),
                    lineno=init_settings.called_line,
                )

        self.pyproject_toml_path = init_settings.pyproject_toml_path
        self._setup_class_vars_from_pyproject_toml_or_by_default()
        self._setup_config_dict_variables_from_pyproject_toml_or_by_default()
        ################ END PYPROJECT ################

        if self.read_config:
            _kwargs["_read_config"] = self.read_config
            self._setup_params(**_kwargs)

        self.arfi_debug = init_settings.init_params.arfi_debug
        return values

    def _setup_read_config(
        self,
        _read_config: bool = None,
        _read_config_force: bool = None,
    ):
        """Setup read_config."""

        if _read_config is not None:
            self.read_config = _read_config
        if _read_config_force is not None:
            self.read_config_force = _read_config_force
        if self.read_config_force is not None:
            self.read_config = self.read_config_force

    def _need_read_pyproject(
        self,
        _read_config: bool = None,
        _read_config_force: bool = None,
    ) -> bool:
        """Checking whether the pyproject.toml file needs to be read."""

        need_read = self.read_pyproject_toml
        read_config_force = self.read_config_force

        if _read_config is not None:
            read_config_force = _read_config
        if _read_config_force is not None:
            read_config_force = _read_config_force
        if read_config_force is not None:
            need_read = read_config_force
        # from handler read_pyproject_toml_force always False
        if self.read_pyproject_toml_force is not None:
            need_read = self.read_pyproject_toml_force
        return need_read

    def _setup_params_from_handler(
        self,
        values: dict[str, Any],
        _mode_dir_inherit_parent: bool = None,
        _file_config_inherit_parent: bool = None,
        _env_config_inherit_parent: bool = None,
        _ordered_settings_inherit_parent: bool = None,
        _handler_inherit_parent: bool = None,
    ) -> None:
        """Setup params from handler."""

        if _handler_tree := values.get("_handler_tree"):
            self.handler_tree = _handler_tree
        if _handler_mode_dir_attr := values.get("_handler_mode_dir_attr"):
            self.mode_dir_attr = _handler_mode_dir_attr
        if _handler_parent_mode_dir := values.get("_handler_parent_mode_dir"):
            self.handler_parent_mode_dir = _handler_parent_mode_dir
            parent_mode_dir = Path(*_handler_parent_mode_dir).as_posix().strip(".")
            self.parent_mode_dir = parent_mode_dir or DEFAULT_PATH_SENTINEL
        if _handler_ordered_settings := values.get("_handler_ordered_settings"):
            if self.ordered_settings_inherit_parent and _ordered_settings_inherit_parent is not False:
                self.ordered_settings = _handler_ordered_settings
        if _handler_parent_file_config := values.get("_handler_parent_file_config"):
            if self.file_config_inherit_parent and _file_config_inherit_parent is not False:
                self.parent_file_config = _handler_parent_file_config
        if _handler_parent_env_config := values.get("_handler_parent_env_config"):
            if self.env_config_inherit_parent and _env_config_inherit_parent is not False:
                self.parent_env_config = _handler_parent_env_config
        _handler_read_pyproject_toml_force = values.get("_handler_read_pyproject_toml_force")
        if _handler_read_pyproject_toml_force is not None:
            self.read_pyproject_toml_force = _handler_read_pyproject_toml_force
        _handler_search_base_dir = values.get("_handler_search_base_dir")
        if _handler_search_base_dir is not None:
            self.search_base_dir = _handler_search_base_dir
        if _handler_main_handler := values.get("_handler_main_handler"):
            if self.handler_inherit_parent and _handler_inherit_parent is not False:
                self.handler = _handler_main_handler

    def _setup_class_vars_from_pyproject_toml_or_by_default(self) -> None:
        """Re-reads variables depending on the current path of the pyproject.toml file."""

        pyproject_or_default_fields = init_settings.init_params.model_fields
        # Fields explicitly set by the user in the class
        modified_class_vars = self.instance._modified_class_vars
        # for field_name in self.model_fields.keys():
        for field_name in self.model_fields.keys():
            if field_name in pyproject_or_default_fields:
                instance_field_name = f"_{field_name}"
                if hasattr(self.instance, instance_field_name):
                    if field_name not in modified_class_vars:
                        field_value = getattr(init_settings.init_params, field_name)
                    else:
                        field_value = getattr(self.instance, instance_field_name)
                    if field_name == "handler":
                        if field_value != self.handler and self.handler != "default_main_handler":
                            field_value = self.handler
                    setattr(self, field_name, field_value)

        self.base_dir = init_settings.base_dir
        self.root_dir = init_settings.root_dir
        if not self.base_dir and not self.search_base_dir and self.instance.BASE_DIR:
            self.base_dir = self.instance.BASE_DIR

    def _setup_config_dict_variables_from_pyproject_toml_or_by_default(self) -> None:
        """Re-reads variables depending on the current path of the pyproject.toml file."""

        # Always reset the settings
        self.class_settings_config = SettingsConfigSchema()
        self.class_file_config = FileConfigSchema()
        self.class_env_config = EnvConfigSchema()

        global_config_keys = list(SettingsConfigSchema.model_fields.keys())
        file_config_keys = list(FileConfigSchema.model_fields.keys())
        env_config_keys = list(EnvConfigSchema.model_fields.keys())
        extra_keys = ["conf_path", "env_path"]
        # setup simple keys
        for key in global_config_keys:
            if key in extra_keys:
                continue
            if key in file_config_keys:
                value = self.file_config.get(key, PydanticUndefined)
                if value is PydanticUndefined:
                    value = self.global_config.get(key, getattr(init_settings.init_params, key))
                setattr(self.class_file_config, key, value)
                setattr(self.class_settings_config, key, value)
            elif key in env_config_keys:
                value = self.env_config.get(key, PydanticUndefined)
                if value is PydanticUndefined:
                    value = self.global_config.get(key, getattr(init_settings.init_params, key))
                setattr(self.class_env_config, key, value)
                setattr(self.class_settings_config, key, value)
            else:
                value = self.global_config.get(key, getattr(init_settings.init_params, key))
                setattr(self.class_settings_config, key, value)

        # SETUP COMMON VALUES
        # ENCODING
        encoding = self.global_config.get(model.ENCODING, init_settings.init_params.encoding)
        self.class_settings_config.encoding = encoding
        conf_file_encoding = self.file_config.get(conf.FILE_ENCODING, PydanticUndefined)
        if conf_file_encoding is PydanticUndefined:
            conf_file_encoding = self.global_config.get(conf.FILE_ENCODING, PydanticUndefined)
        if conf_file_encoding is PydanticUndefined:
            conf_file_encoding = init_settings.init_params.conf_file_encoding or encoding
        self.class_file_config.conf_file_encoding = conf_file_encoding
        self.class_settings_config.conf_file_encoding = conf_file_encoding

        env_file_encoding = self.env_config.get(env.FILE_ENCODING, PydanticUndefined)
        if env_file_encoding is PydanticUndefined:
            env_file_encoding = self.global_config.get(env.FILE_ENCODING, PydanticUndefined)
        if env_file_encoding is PydanticUndefined:
            env_file_encoding = init_settings.init_params.env_file_encoding or encoding
        self.class_env_config.env_file_encoding = env_file_encoding
        self.class_settings_config.env_file_encoding = env_file_encoding

        # CASE SENSITIVE
        case_sensitive = self.global_config.get(model.CASE_SENSITIVE, init_settings.init_params.case_sensitive)
        self.class_settings_config.case_sensitive = case_sensitive
        conf_case_sensitive = self.file_config.get(conf.CASE_SENSITIVE, PydanticUndefined)
        if conf_case_sensitive is PydanticUndefined:
            conf_case_sensitive = self.global_config.get(conf.CASE_SENSITIVE, PydanticUndefined)
        if conf_case_sensitive is PydanticUndefined:
            conf_case_sensitive = init_settings.init_params.conf_case_sensitive
        if conf_case_sensitive is None:
            conf_case_sensitive = case_sensitive
        self.class_file_config.conf_case_sensitive = conf_case_sensitive
        self.class_settings_config.conf_case_sensitive = conf_case_sensitive

        env_case_sensitive = self.env_config.get(env.CASE_SENSITIVE, PydanticUndefined)
        if env_case_sensitive is PydanticUndefined:
            env_case_sensitive = self.global_config.get(env.CASE_SENSITIVE, PydanticUndefined)
        if env_case_sensitive is PydanticUndefined:
            env_case_sensitive = init_settings.init_params.env_case_sensitive
        if env_case_sensitive is None:
            env_case_sensitive = case_sensitive
        self.class_env_config.env_case_sensitive = env_case_sensitive
        self.class_settings_config.env_case_sensitive = env_case_sensitive

        # IGNORE MISSING
        ignore_missing = self.global_config.get(model.IGNORE_MISSING, init_settings.init_params.ignore_missing)
        self.class_settings_config.ignore_missing = ignore_missing
        conf_ignore_missing = self.file_config.get(conf.IGNORE_MISSING, PydanticUndefined)
        if conf_ignore_missing is PydanticUndefined:
            conf_ignore_missing = self.global_config.get(conf.IGNORE_MISSING, PydanticUndefined)
        if conf_ignore_missing is PydanticUndefined:
            conf_ignore_missing = init_settings.init_params.conf_ignore_missing
        if conf_ignore_missing is None:
            conf_ignore_missing = ignore_missing
        self.class_file_config.conf_ignore_missing = conf_ignore_missing
        self.class_settings_config.conf_ignore_missing = conf_ignore_missing

        env_ignore_missing = self.env_config.get(env.IGNORE_MISSING, PydanticUndefined)
        if env_ignore_missing is PydanticUndefined:
            env_ignore_missing = self.global_config.get(env.IGNORE_MISSING, PydanticUndefined)
        if env_ignore_missing is PydanticUndefined:
            env_ignore_missing = init_settings.init_params.env_ignore_missing
        if env_ignore_missing is None:
            env_ignore_missing = ignore_missing
        self.class_env_config.env_ignore_missing = env_ignore_missing
        self.class_settings_config.env_ignore_missing = env_ignore_missing

        # check custom_ext_handler
        if not self.class_file_config.conf_ext or self.class_file_config.conf_ext == LIST_STR_SENTINEL:
            try:
                assert (
                    self.class_file_config.conf_custom_ext_handler
                ), "`conf_custom_ext_handler` must be defined if conf_ext is empty"
            except AssertionError as e:
                raise ArFiSettingsError(e) from e

        # setup include and exclude inherit parent params
        include_inherit_parent = self.class_settings_config.include_inherit_parent or []
        exclude_inherit_parent = self.class_settings_config.exclude_inherit_parent or []
        conf_include_inherit_parent = self.class_file_config.conf_include_inherit_parent or []
        conf_exclude_inherit_parent = self.class_file_config.conf_exclude_inherit_parent or []
        env_include_inherit_parent = self.class_env_config.env_include_inherit_parent or []
        env_exclude_inherit_parent = self.class_env_config.env_exclude_inherit_parent or []

        class_include_inherit_parent = []
        class_exclude_inherit_parent = []
        class_conf_include_inherit_parent = []
        class_conf_exclude_inherit_parent = []
        class_env_include_inherit_parent = []
        class_env_exclude_inherit_parent = []

        if not conf_exclude_inherit_parent:
            for e in exclude_inherit_parent:
                if e in CONF_INCLUDE_EXLUDE_PARAMS:
                    class_conf_exclude_inherit_parent.append(e)
        else:
            class_conf_exclude_inherit_parent.extend(conf_exclude_inherit_parent)

        if not env_exclude_inherit_parent:
            for e in exclude_inherit_parent:
                if e in ENV_INCLUDE_EXLUDE_PARAMS:
                    class_env_exclude_inherit_parent.append(e)
        else:
            class_env_exclude_inherit_parent.extend(env_exclude_inherit_parent)

        class_exclude_inherit_parent.extend(class_conf_exclude_inherit_parent)
        class_exclude_inherit_parent.extend(class_env_exclude_inherit_parent)

        for ci in conf_include_inherit_parent:
            if ci not in class_conf_exclude_inherit_parent:
                if class_conf_exclude_inherit_parent and ci in class_conf_exclude_inherit_parent:
                    continue
                class_conf_include_inherit_parent.append(ci)

        if not class_conf_include_inherit_parent:
            for i in include_inherit_parent:
                if i in CONF_INCLUDE_EXLUDE_PARAMS:
                    if class_conf_exclude_inherit_parent and i in class_conf_exclude_inherit_parent:
                        continue
                    class_conf_include_inherit_parent.append(i)

        for ei in env_include_inherit_parent:
            if ei not in class_env_exclude_inherit_parent:
                if class_env_exclude_inherit_parent and ei in class_env_exclude_inherit_parent:
                    continue
                class_env_include_inherit_parent.append(ei)

        if not class_env_include_inherit_parent:
            for i in include_inherit_parent:
                if i in ENV_INCLUDE_EXLUDE_PARAMS:
                    if class_env_exclude_inherit_parent and i in class_env_exclude_inherit_parent:
                        continue
                    class_env_include_inherit_parent.append(i)

        class_include_inherit_parent.extend(class_conf_include_inherit_parent)
        class_include_inherit_parent.extend(class_env_include_inherit_parent)
        self.class_file_config.conf_include_inherit_parent = class_conf_include_inherit_parent
        self.class_file_config.conf_exclude_inherit_parent = class_conf_exclude_inherit_parent
        self.class_env_config.env_include_inherit_parent = class_env_include_inherit_parent
        self.class_env_config.env_exclude_inherit_parent = class_env_exclude_inherit_parent
        self.class_settings_config.conf_include_inherit_parent = class_conf_include_inherit_parent
        self.class_settings_config.conf_exclude_inherit_parent = class_conf_exclude_inherit_parent
        self.class_settings_config.env_include_inherit_parent = class_env_include_inherit_parent
        self.class_settings_config.env_exclude_inherit_parent = class_env_exclude_inherit_parent
        self.class_settings_config.include_inherit_parent = class_include_inherit_parent
        self.class_settings_config.exclude_inherit_parent = class_exclude_inherit_parent

        self.settings_config = self.class_settings_config

    def _setup_instance_and_default_value(self, instance: "ArFiSettings", **values) -> None:
        """Setup default value from class."""

        if self.instance is not None:
            return
        # setup class constants
        self.instance = instance
        if isinstance(instance.handler, str):
            self.handler = instance.handler
        self.handler_inherit_parent = instance.handler_inherit_parent
        self.handler_class = instance.handler_class
        self.read_config = instance.read_config
        self.read_config_force = instance.read_config_force
        self.read_pyproject_toml = instance.read_pyproject_toml
        self.mode_dir = instance._mode_dir
        self.search_base_dir = instance._search_base_dir
        self.mode_dir_attr = instance._mode_dir_attr

        # setup class config
        self.file_config = instance.file_config
        self.env_config = instance.env_config
        self.global_config = GlobalConfigDict()
        for key, value in instance.model_config.items():
            if key in list(init_settings.init_params.model_fields.keys()):
                self.global_config[key] = value

        # setup default value
        self.values_by_default = dict(values) or {}

        # search nested mode_dir
        for base in type(instance).__bases__:
            if hasattr(base, "mode_dir"):
                self.nested_mode_dir = getattr(base, "mode_dir")
                break

    def _setup_params(
        self,
        _read_config: bool = None,
        _read_config_force: bool = None,
        _mode_dir: PathType | None = DEFAULT_PATH_SENTINEL,
        _mode_dir_inherit_nested: bool = None,
        _mode_dir_inherit_parent: bool = None,
        _file_config_inherit_parent: bool = None,
        _env_config_inherit_parent: bool = None,
        _conf_file: MultiPathType | None = DEFAULT_PATH_SENTINEL,
        _conf_dir: MultiPathType | None = DEFAULT_PATH_SENTINEL,
        _conf_ext: str | list[str, ...] = [],
        _conf_file_encoding: str | None = STR_SENTINEL,
        _conf_case_sensitive: bool = None,
        _conf_ignore_missing: bool = None,
        _conf_custom_ext_handler: str | dict[str, str] | None = STR_SENTINEL,
        _conf_include_inherit_parent: list[Literal[*CONF_INCLUDE_EXLUDE_PARAMS], ...] = None,
        _conf_exclude_inherit_parent: list[Literal[*CONF_INCLUDE_EXLUDE_PARAMS], ...] = None,
        _env_file: MultiPathType | None = DEFAULT_PATH_SENTINEL,
        _env_prefix: str = None,
        _env_prefix_as_mode_dir: bool = None,
        _env_prefix_as_nested_mode_dir: bool = None,
        _env_prefix_as_source_mode_dir: bool = None,
        _env_file_encoding: str | None = STR_SENTINEL,
        _env_case_sensitive: bool = None,
        _env_nested_delimiter: str = None,
        _env_ignore_missing: bool = None,
        _env_include_inherit_parent: list[Literal[*ENV_INCLUDE_EXLUDE_PARAMS], ...] = None,
        _env_exclude_inherit_parent: list[Literal[*ENV_INCLUDE_EXLUDE_PARAMS], ...] = None,
        _case_sensitive: bool = None,
        _ignore_missing: bool = None,
        _encoding: str | None = STR_SENTINEL,
        _secrets_dir: PathType | None = DEFAULT_PATH_SENTINEL,
        _include_inherit_parent: list[Literal[*INCLUDE_EXLUDE_PARAMS], ...] = None,
        _exclude_inherit_parent: list[Literal[*INCLUDE_EXLUDE_PARAMS], ...] = None,
        _handler_inherit_parent: bool = None,
        _ordered_settings: list[str, ...] = LIST_STR_SENTINEL,
        _ordered_settings_inherit_parent: bool = None,
        _cli: bool = None,
    ):
        """Load config params."""

        if _mode_dir != DEFAULT_PATH_SENTINEL:
            self.mode_dir = _mode_dir
        if _mode_dir_inherit_nested is not None:
            self.mode_dir_inherit_nested = _mode_dir_inherit_nested
        if _mode_dir_inherit_parent is not None:
            self.mode_dir_inherit_parent = _mode_dir_inherit_parent
        if _file_config_inherit_parent is not None:
            self.file_config_inherit_parent = _file_config_inherit_parent
        if _env_config_inherit_parent is not None:
            self.env_config_inherit_parent = _env_config_inherit_parent
        if _handler_inherit_parent is not None:
            self.handler_inherit_parent = _handler_inherit_parent
        if _ordered_settings != LIST_STR_SENTINEL:
            self.ordered_settings = _ordered_settings
        if _ordered_settings_inherit_parent is not None:
            self.ordered_settings_inherit_parent = _ordered_settings_inherit_parent

        if _case_sensitive is not None:
            if _conf_case_sensitive is None:
                _conf_case_sensitive = _case_sensitive
            if _env_case_sensitive is None:
                _env_case_sensitive = _case_sensitive

        if _ignore_missing is not None:
            if _conf_ignore_missing is None:
                _conf_ignore_missing = _ignore_missing
            if _env_ignore_missing is None:
                _env_ignore_missing = _ignore_missing

        if _encoding is not STR_SENTINEL:
            if _conf_file_encoding is STR_SENTINEL:
                _conf_file_encoding = _encoding
            if _env_file_encoding is STR_SENTINEL:
                _env_file_encoding = _encoding

        exclude_inherit_parent = self.settings_init_params.exclude_inherit_parent
        include_inherit_parent = self.settings_init_params.include_inherit_parent

        if not _env_exclude_inherit_parent and exclude_inherit_parent:
            _env_exclude_inherit_parent = []
            for ee in exclude_inherit_parent:
                if ee in ENV_INCLUDE_EXLUDE_PARAMS:
                    _env_exclude_inherit_parent.append(ee)

        if not _env_include_inherit_parent and include_inherit_parent:
            _env_include_inherit_parent = []
            for ei in include_inherit_parent:
                if ei in ENV_INCLUDE_EXLUDE_PARAMS:
                    if _env_exclude_inherit_parent and ei in _env_exclude_inherit_parent:
                        continue
                    _env_include_inherit_parent.append(ei)

        if not _conf_exclude_inherit_parent and exclude_inherit_parent:
            _conf_exclude_inherit_parent = []
            for ce in exclude_inherit_parent:
                if ce in CONF_INCLUDE_EXLUDE_PARAMS:
                    _conf_exclude_inherit_parent.append(ce)

        if not _conf_include_inherit_parent and include_inherit_parent:
            _conf_include_inherit_parent = []
            for ci in include_inherit_parent:
                if ci in CONF_INCLUDE_EXLUDE_PARAMS:
                    if _conf_exclude_inherit_parent and ci in _conf_exclude_inherit_parent:
                        continue
                    _conf_include_inherit_parent.append(ci)

        self.init_file_config = FileConfigSchema(
            conf_file=_conf_file,
            conf_dir=_conf_dir,
            conf_ext=_conf_ext,
            conf_file_encoding=_conf_file_encoding,
            conf_case_sensitive=_conf_case_sensitive,
            conf_ignore_missing=_conf_ignore_missing,
            conf_custom_ext_handler=_conf_custom_ext_handler,
            conf_include_inherit_parent=_conf_include_inherit_parent,
            conf_exclude_inherit_parent=_conf_exclude_inherit_parent,
        )
        self.init_env_config = EnvConfigSchema(
            env_file=_env_file,
            env_prefix=_env_prefix,
            env_prefix_as_mode_dir=_env_prefix_as_mode_dir,
            env_prefix_as_nested_mode_dir=_env_prefix_as_nested_mode_dir,
            env_prefix_as_source_mode_dir=_env_prefix_as_source_mode_dir,
            env_file_encoding=_env_file_encoding,
            env_case_sensitive=_env_case_sensitive,
            env_nested_delimiter=_env_nested_delimiter,
            env_ignore_missing=_env_ignore_missing,
            env_include_inherit_parent=_env_include_inherit_parent,
            env_exclude_inherit_parent=_env_exclude_inherit_parent,
        )
        if self.init_file_config.conf_include_inherit_parent or self.init_env_config.env_include_inherit_parent:
            self.settings_init_params.include_inherit_parent = []
            if self.init_file_config.conf_include_inherit_parent:
                self.settings_init_params.conf_include_inherit_parent = (
                    self.init_file_config.conf_include_inherit_parent
                )
                self.settings_init_params.include_inherit_parent.extend(
                    self.init_file_config.conf_include_inherit_parent
                )
            if self.init_env_config.env_include_inherit_parent:
                self.settings_init_params.env_include_inherit_parent = self.init_env_config.env_include_inherit_parent
                self.settings_init_params.include_inherit_parent.extend(self.init_env_config.env_include_inherit_parent)

        if self.init_file_config.conf_exclude_inherit_parent or self.init_env_config.env_exclude_inherit_parent:
            self.settings_init_params.exclude_inherit_parent = []
            if self.init_file_config.conf_exclude_inherit_parent:
                self.settings_init_params.conf_exclude_inherit_parent = (
                    self.init_file_config.conf_exclude_inherit_parent
                )
                self.settings_init_params.exclude_inherit_parent.extend(
                    self.init_file_config.conf_exclude_inherit_parent
                )
            if self.init_env_config.env_exclude_inherit_parent:
                self.settings_init_params.env_exclude_inherit_parent = self.init_env_config.env_exclude_inherit_parent
                self.settings_init_params.exclude_inherit_parent.extend(self.init_env_config.env_exclude_inherit_parent)

        # Very important place !!!
        # here computes and updates the main settings config
        self._compute_include_exclude_inherit_parent()  # Important: must run before computes file and env config
        self.computed_file_config = self.instance_computed_file_config
        self.computed_env_config = self.instance_computed_env_config

        self.settings_config.update(**self.computed_file_config)
        self.settings_config.update(**self.computed_env_config)

        self.mode_dir_path = self.instance_mode_dir_path
        self.computed_mode_dir = self.instance_computed_mode_dir
        if _case_sensitive is not None:
            self.settings_config.case_sensitive = _case_sensitive
        if _ignore_missing is not None:
            self.settings_config.ignore_missing = _ignore_missing
        if _encoding is not STR_SENTINEL:
            self.settings_config.encoding = _encoding
        if _secrets_dir != DEFAULT_PATH_SENTINEL:
            self.settings_config.secrets_dir = _secrets_dir
        if _cli is not None:
            self.settings_config.cli = _cli
        self.conf_path = self.instance_conf_path
        self.env_path = self.instance_env_path
        self.settings_config.update(
            **dict(
                conf_path=self.conf_path,
                env_path=self.env_path,
            )
        )

    @property
    def instance_mode_dir_path(self) -> PathType:
        """Generate path to mode directory."""

        if self.parent_mode_dir and self.parent_mode_dir != DEFAULT_PATH_SENTINEL and self.mode_dir_inherit_parent:
            if self.nested_mode_dir and self.nested_mode_dir != DEFAULT_PATH_SENTINEL and self.mode_dir_inherit_nested:
                return Path(self.parent_mode_dir, self.nested_mode_dir, self.mode_dir or "")
            return Path(self.parent_mode_dir, self.mode_dir or "")
        elif self.nested_mode_dir and self.nested_mode_dir != DEFAULT_PATH_SENTINEL and self.mode_dir_inherit_nested:
            return Path(self.nested_mode_dir, self.mode_dir or "")
        else:
            return Path(self.mode_dir or "")

    @property
    def instance_computed_mode_dir(self) -> str:
        return self.instance_mode_dir_path.as_posix().strip(".")

    @property
    def instance_computed_nested_mode_dir(self) -> str:
        """Compute nested_mode_dir."""

        if self.nested_mode_dir and self.nested_mode_dir != DEFAULT_PATH_SENTINEL and self.mode_dir_inherit_nested:
            return Path(self.nested_mode_dir, self.mode_dir or "").as_posix().strip(".")
        else:
            return Path(self.mode_dir or "").as_posix().strip(".")

    @property
    def instance_source_mode_dir(self) -> str:
        if isinstance(self.mode_dir, Path):
            return self.mode_dir.as_posix()
        return self.mode_dir or ""

    @property
    def instance_conf_path(self) -> list[Path]:
        """Create full config path from file and dir settings."""

        conf_path = []
        conf_file = self.computed_file_config.get(conf.FILE)
        if not conf_file:
            return conf_path
        if isinstance(conf_file, (str, Path)):
            conf_file = [Path(conf_file)]
        try:
            assert isinstance(
                conf_file, (list, tuple)
            ), f"{self.instance.__class__.__name__}.file_config['{conf.FILE}'] type must be a arfi_settings.MultiPathType"
        except AssertionError as e:
            raise ArFiSettingsError(e) from e

        conf_dir = self.computed_file_config.get(conf.DIR) or ""
        if isinstance(conf_dir, (str, Path)):
            conf_dir = [Path(conf_dir)]
        try:
            assert isinstance(
                conf_dir, (list, tuple)
            ), f"{self.instance.__class__.__name__}.file_config['{conf.DIR}'] type must be a arfi_settings.MultiPathType"
        except AssertionError as e:
            raise ArFiSettingsError(e) from e

        for directory in conf_dir:
            directory = Path(directory).expanduser()
            for file in conf_file:
                file = Path(file).expanduser()
                path = Path(directory, self.instance_mode_dir_path, file).expanduser()
                conf_path.append(path)
        return conf_path

    @property
    def instance_env_path(self) -> list[Path]:
        """Create full env path from file and dir settings."""

        env_path = []
        env_file = self.computed_env_config.get(env.FILE)
        if not env_file:
            return env_path
        if isinstance(env_file, (str, Path)):
            env_file = [Path(env_file)]
        try:
            assert isinstance(
                env_file, (list, tuple)
            ), f"{self.instance.__class__.__name__}.env_config['{env.FILE}'] must be a arfi_settings.MultiPathType"
        except AssertionError as e:
            raise ArFiSettingsError(e) from e

        for file in env_file:
            path = Path(file).expanduser()
            env_path.append(path)
        return env_path

    @property
    def instance_computed_file_config(self) -> FileConfigDict:
        """Compute file_config."""

        file_config = copy.deepcopy(self.class_file_config)
        file_config.conf_include_inherit_parent = self.settings_config.conf_include_inherit_parent
        file_config.conf_exclude_inherit_parent = self.settings_config.conf_exclude_inherit_parent
        if self.file_config_inherit_parent:
            parent_file_config = self.parent_file_config.model_dump(exclude_defaults=True)
            if file_config.conf_include_inherit_parent:
                for key, value in parent_file_config.items():
                    if key in file_config.conf_include_inherit_parent:
                        if key not in ["conf_include_inherit_parent", "conf_exclude_inherit_parent"]:
                            setattr(file_config, key, value)
                            self.inherited_params.append(key)
            elif file_config.conf_exclude_inherit_parent:
                for key, value in parent_file_config.items():
                    if key not in file_config.conf_exclude_inherit_parent:
                        if key not in ["conf_include_inherit_parent", "conf_exclude_inherit_parent"]:
                            setattr(file_config, key, value)
                            self.inherited_params.append(key)
            else:
                for key, value in parent_file_config.items():
                    if key not in ["conf_include_inherit_parent", "conf_exclude_inherit_parent"]:
                        setattr(file_config, key, value)
                        self.inherited_params.append(key)

        init_file_config = self.init_file_config.model_dump(exclude_defaults=True)
        for key, value in init_file_config.items():
            if key not in ["conf_include_inherit_parent", "conf_exclude_inherit_parent"]:
                setattr(file_config, key, value)

        return file_config.conf_dict

    @property
    def instance_computed_env_config(self) -> EnvConfigDict:
        """Compute env_config."""

        env_config = copy.deepcopy(self.class_env_config)
        env_config.env_include_inherit_parent = self.settings_config.env_include_inherit_parent
        env_config.env_exclude_inherit_parent = self.settings_config.env_exclude_inherit_parent
        if self.env_config_inherit_parent:
            parent_env_config = self.parent_env_config.model_dump(exclude_defaults=True)
            if env_config.env_include_inherit_parent:
                for key, value in parent_env_config.items():
                    if key in env_config.env_include_inherit_parent:
                        if key not in ["env_include_inherit_parent", "env_exclude_inherit_parent"]:
                            setattr(env_config, key, value)
                            self.inherited_params.append(key)
            elif env_config.env_exclude_inherit_parent:
                for key, value in parent_env_config.items():
                    if key not in env_config.env_exclude_inherit_parent:
                        if key not in ["env_include_inherit_parent", "env_exclude_inherit_parent"]:
                            setattr(env_config, key, value)
                            self.inherited_params.append(key)
            else:
                for key, value in parent_env_config.items():
                    if key not in ["env_include_inherit_parent", "env_exclude_inherit_parent"]:
                        setattr(env_config, key, value)
                        self.inherited_params.append(key)

        init_env_config = self.init_env_config.model_dump(exclude_defaults=True)
        for key, value in init_env_config.items():
            if key not in ["env_include_inherit_parent", "env_exclude_inherit_parent"]:
                setattr(env_config, key, value)

        # search env prefix
        env_prefix = env_config.env_prefix
        found_env_prefix = ""
        if env_config.env_prefix_as_mode_dir:
            found_env_prefix = self.instance_computed_mode_dir.replace("/", "_").strip(".")
        if env_config.env_prefix_as_nested_mode_dir:
            found_env_prefix = self.instance_computed_nested_mode_dir.replace("/", "_").strip(".")
        if env_config.env_prefix_as_source_mode_dir:
            found_env_prefix = self.instance_source_mode_dir.replace("/", "_").strip(".")
        if found_env_prefix:
            env_prefix = f"{found_env_prefix}_"
        env_config.env_prefix = env_prefix

        return env_config.env_dict

    def _compute_include_exclude_inherit_parent(self) -> None:
        """Compute include and exclude inherit parent params and setup it to settings_config."""

        result_exclude = []
        result_include = []
        result_conf_exclude = []
        result_conf_include = []
        result_env_exclude = []
        result_env_include = []
        init_exclude = self.settings_init_params.exclude_inherit_parent
        init_include = self.settings_init_params.include_inherit_parent

        # ######### FILE EXCLUDE #########
        class_conf_exclude = self.class_file_config.conf_exclude_inherit_parent or []
        result_conf_exclude = class_conf_exclude
        parent_conf_exclude = self.parent_file_config.conf_exclude_inherit_parent or []
        init_conf_exclude = self.init_file_config.conf_exclude_inherit_parent
        if self.file_config_inherit_parent:
            aux_exclude = result_conf_exclude
            if init_conf_exclude:
                aux_exclude = init_conf_exclude
            elif init_exclude:
                res_exclude = []
                for ie in init_exclude:
                    if ie in CONF_INCLUDE_EXLUDE_PARAMS:
                        res_exclude.append(ie)
                if res_exclude:
                    aux_exclude = res_exclude
            if conf.CONF_EXCLUDE_INHERIT_PARENT not in aux_exclude and parent_conf_exclude:
                result_conf_exclude = parent_conf_exclude
        if init_conf_exclude is not None:
            result_conf_exclude = init_conf_exclude
        elif init_exclude is not None:
            res_exclude = []
            for ie in init_exclude:
                if ie in CONF_INCLUDE_EXLUDE_PARAMS:
                    res_exclude.append(ie)
            if res_exclude:
                result_conf_exclude = res_exclude
        self.settings_config.conf_exclude_inherit_parent = result_conf_exclude
        if self.file_config_inherit_parent:
            for key in ["conf_include_inherit_parent", "conf_exclude_inherit_parent"]:
                if key not in result_conf_exclude and key not in self.inherited_params:
                    self.inherited_params.append(key)

        # ######### FILE INCLUDE #########
        class_conf_include = self.class_file_config.conf_include_inherit_parent or []
        result_conf_include = class_conf_include
        parent_conf_include = self.parent_file_config.conf_include_inherit_parent or []
        init_conf_include = self.init_file_config.conf_include_inherit_parent
        if self.file_config_inherit_parent:
            if conf.CONF_INCLUDE_INHERIT_PARENT not in result_conf_exclude and parent_conf_include:
                result_conf_include = []
                for pci in parent_conf_include:
                    if pci not in result_conf_exclude:
                        result_conf_include.append(pci)
        if init_conf_include is not None:
            result_conf_include = []
            for ici in init_conf_include:
                if ici not in result_conf_exclude:
                    result_conf_include.append(ici)
        elif init_include is not None:
            res_include = []
            for ii in init_include:
                if ii in CONF_INCLUDE_EXLUDE_PARAMS:
                    res_include.append(ii)
            if res_include:
                result_conf_include = res_include
        aux_res = result_conf_include
        result_conf_include = []
        for rci in aux_res:
            if rci not in result_conf_exclude:
                result_conf_include.append(rci)
        self.settings_config.conf_include_inherit_parent = result_conf_include

        # ######### ENV EXCLUDE #########
        class_env_exclude = self.class_env_config.env_exclude_inherit_parent or []
        result_env_exclude = class_env_exclude
        parent_env_exclude = self.parent_env_config.env_exclude_inherit_parent or []
        init_env_exclude = self.init_env_config.env_exclude_inherit_parent
        if self.env_config_inherit_parent:
            aux_exclude = result_env_exclude
            if init_env_exclude:
                aux_exclude = init_env_exclude
            elif init_exclude:
                res_exclude = []
                for ie in init_exclude:
                    if ie in CONF_INCLUDE_EXLUDE_PARAMS:
                        res_exclude.append(ie)
                if res_exclude:
                    aux_exclude = res_exclude
            if env.ENV_EXCLUDE_INHERIT_PARENT not in aux_exclude and parent_env_exclude:
                result_env_exclude = parent_env_exclude
        if init_env_exclude is not None:
            result_env_exclude = init_env_exclude
        elif init_exclude is not None:
            res_exclude = []
            for ie in init_exclude:
                if ie in ENV_INCLUDE_EXLUDE_PARAMS:
                    res_exclude.append(ie)
            if res_exclude:
                result_env_exclude = res_exclude
        self.settings_config.env_exclude_inherit_parent = result_env_exclude
        if self.env_config_inherit_parent:
            for key in ["env_include_inherit_parent", "env_exclude_inherit_parent"]:
                if key not in result_env_exclude and key not in self.inherited_params:
                    self.inherited_params.append(key)

        # ######### ENV INCLUDE #########
        class_env_include = self.class_env_config.env_include_inherit_parent or []
        result_env_include = class_env_include
        parent_env_include = self.parent_env_config.env_include_inherit_parent or []
        init_env_include = self.init_env_config.env_include_inherit_parent
        if self.env_config_inherit_parent:
            if env.ENV_INCLUDE_INHERIT_PARENT not in result_env_exclude and parent_env_include:
                result_env_include = []
                for pei in parent_env_include:
                    if pei not in result_env_exclude:
                        result_env_include.append(pei)
        if init_env_include is not None:
            result_env_include = []
            for iei in init_env_include:
                if iei not in result_env_exclude:
                    result_env_include.append(iei)
        elif init_include is not None:
            res_include = []
            for ii in init_include:
                if ii in ENV_INCLUDE_EXLUDE_PARAMS:
                    res_include.append(ii)
            if res_include:
                result_env_include = res_include

        aux_res = result_env_include
        result_env_include = []
        for rei in aux_res:
            if rei not in result_env_exclude:
                result_env_include.append(rei)
        self.settings_config.env_include_inherit_parent = result_env_include

        # ######### RESULT EXCLUDE #########
        result_include.extend(result_conf_include)
        result_include.extend(result_env_include)
        result_exclude.extend(result_conf_exclude)
        result_exclude.extend(result_env_exclude)
        self.settings_config.include_inherit_parent = result_include
        self.settings_config.exclude_inherit_parent = result_exclude

        DEBUG_DICT = {
            "COMMON_INIT_CONFIG": {
                "init_exclude": init_exclude,
                "init_include": init_include,
            },
            "FILE_CONFIG": {
                "exclude": {
                    "class_conf_exclude": class_conf_exclude,
                    "parent_conf_exclude": parent_conf_exclude,
                    "init_conf_exclude": init_conf_exclude,
                    "result_conf_exclude": result_conf_exclude,
                },
                "include": {
                    "class_conf_include": class_conf_include,
                    "parent_conf_include": parent_conf_include,
                    "init_conf_include": init_conf_include,
                    "result_conf_include": result_conf_include,
                },
            },
            "ENV_CONFIG": {
                "exclude": {
                    "class_env_exclude": class_env_exclude,
                    "parent_env_exclude": parent_env_exclude,
                    "init_env_exclude": init_env_exclude,
                    "result_env_exclude": result_env_exclude,
                },
                "include": {
                    "class_env_include": class_env_include,
                    "parent_env_include": parent_env_include,
                    "init_env_include": init_env_include,
                    "result_env_include": result_env_include,
                },
            },
            "RESULT_CONFIG": {
                "result_include": result_include,
                "result_exclude": result_exclude,
            },
        }
        if self.arfi_dev_debug:
            # TODO: add debug mode
            import pprint

            pprint.pprint(DEBUG_DICT)

    def extract(self, instance: "ArFiSettings") -> None:
        """Setup config to instance."""

        for field_name in self.model_fields.keys():
            if field_name in self._class_vars:
                continue
            instance_field_name = f"_{field_name}"
            setattr(instance, instance_field_name, getattr(self, field_name))


class ConfigStorage:
    storage: dict[int, InstanceConfig] = dict()

    def get_config(self, instance: "ArFiSettings") -> InstanceConfig:
        """Create or return config for instance."""

        assert is_settings(type(instance)), f"type(instance) must be a subclass ArFiSettings, got {type(instance)}"
        instance_id = id(instance)
        instance_config = self.storage.get(instance_id, PydanticUndefined)
        if instance_config is PydanticUndefined:
            instance_config = InstanceConfig()
            self.storage[instance_id] = instance_config
        return instance_config


config_storage = ConfigStorage()



import pytest

from arfi_settings import ArFiSettings
from pathlib import PosixPath


# @pytest.mark.current
@pytest.mark.field_default_params
def test_read_config_default_params(simple_data_config_config_toml, path_base_dir):
    """Test default params _read_config and _read_config_force."""

    class AppConfig(ArFiSettings):
        path_config_file: str = "default"

    config = AppConfig()
    assert config.path_config_file == "config/config.toml"
    config = AppConfig(_read_config=False)
    assert config.path_config_file == "default"
    config = AppConfig(_read_config=False, _read_config_force=True)
    assert config.path_config_file == "config/config.toml"

    class AppConfig(ArFiSettings):
        read_config = False
        path_config_file: str = "default"

    config = AppConfig()
    assert config.path_config_file == "default"
    config = AppConfig(_read_config=True)
    assert config.path_config_file == "config/config.toml"

    class AppConfig(ArFiSettings):
        read_config = False
        read_config_force = False
        path_config_file: str = "default"

    config = AppConfig()
    assert config.path_config_file == "default"
    config = AppConfig(_read_config=True)
    assert config.path_config_file == "default"
    config = AppConfig(_read_config_force=True)
    assert config.path_config_file == "config/config.toml"

    class AppConfig(ArFiSettings):
        read_config = True
        read_config_force = False
        path_config_file: str = "default"

    config = AppConfig()
    assert config.path_config_file == "default"
    config = AppConfig(_read_config_force=True)
    assert config.path_config_file == "config/config.toml"

    class AppConfig(ArFiSettings):
        read_config = False
        read_config_force = True
        path_config_file: str = "default"

    config = AppConfig()
    assert config.path_config_file == "config/config.toml"
    config = AppConfig(_read_config_force=False)
    assert config.path_config_file == "default"


# @pytest.mark.current
@pytest.mark.field_default_params
@pytest.mark.mode_dir
def test_mode_dir_default_params(cwd_to_tmp):
    """Test default params _mode_dir."""
    # from pathlib import Path
    #
    # print()
    # print(Path().cwd())

    class AppSettings(ArFiSettings):
        pass

    class AppConfig(ArFiSettings):
        app: AppSettings = AppSettings()
        read_config_force = False

    config = AppConfig()
    assert config.mode_dir == PosixPath(".")
    assert config.app.mode_dir == "app"

    class AppConfig(ArFiSettings):
        app: AppSettings = AppSettings()

    config = AppConfig()
    assert config.mode_dir == PosixPath(".")
    assert config.app.mode_dir == "app"

    class AppConfig(ArFiSettings):
        app: AppSettings = AppSettings(_mode_dir="test_mode_dir")

    config = AppConfig()
    assert config.mode_dir == PosixPath(".")
    assert config.app.mode_dir == "test_mode_dir"


#



import argparse
import sys

import pytest
from pydantic import AliasChoices, Field

from arfi_settings import (
    ArFiReader,
    ArFiSettings,
    SettingsConfigDict,
)


# @pytest.mark.current
@pytest.mark.cli
def test_cli(monkeypatch, cwd_to_tmp):
    def parse_args():
        parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        parser.add_argument(
            "--debug",
            type=bool,
            const=True,
            nargs="?",
            help="Debug application",
        )
        parser.add_argument(
            "-v",
            "--verbose",
            action="count",
            default=0,
        )
        parser.add_argument(
            "-p",
            "--price",
            type=int,
            action="store",
            help="Fruit price",
        )
        parser.add_argument(
            "-n",
            "--name",
            type=str,
            default="default_test_name",
            action="store",
            help="Application name",
        )
        cli_options = parser.parse_args()
        return dict(cli_options._get_kwargs())

    ArFiReader.setup_cli_reader(parse_args)

    class Cost(ArFiSettings):
        DEBUG: bool = False
        price: int
        verbose: int

        model_config = SettingsConfigDict(
            cli=True,
        )

    class Fruit(ArFiSettings):
        DEBUG: bool = False
        cost: Cost
        name: str

    class AppConfig(ArFiSettings):
        DEBUG: bool = False
        name: str
        fruit: Fruit
        model_config = SettingsConfigDict(
            cli=True,
        )

    monkeypatch.setenv("name", "Orange")
    monkeypatch.setattr(sys, "argv", [sys.argv[0], "-vvv", "--debug", "-p", "99", "--name", "SuperApp"])
    config = AppConfig()
    assert config.DEBUG is True
    assert config.fruit.DEBUG is False
    assert config.fruit.cost.DEBUG is True
    assert config.name == "SuperApp"
    assert config.fruit.name == "Orange"
    assert config.fruit.cost.price == 99
    assert config.fruit.cost.verbose == 3

    class AppConfig(ArFiSettings):
        SuperName: str = Field(alias="name")
        model_config = SettingsConfigDict(
            cli=True,
        )

    config = AppConfig()
    assert config.SuperName == "SuperApp"

    monkeypatch.setenv("EnvName", "EnvName")
    monkeypatch.setattr(sys, "argv", [sys.argv[0], "--name", "CliAppName"])

    class AppConfig(ArFiSettings):
        SuperName: str = Field(validation_alias=AliasChoices("EnvName", "name"))
        model_config = SettingsConfigDict(
            cli=True,
        )

    config = AppConfig()
    assert config.SuperName == "CliAppName"

    class AppConfig(ArFiSettings):
        SuperName: str = Field(validation_alias=AliasChoices("EnvName", "name"))
        model_config = SettingsConfigDict(
            cli=True,
        )
        ordered_settings = [
            "env",
            "cli",
        ]

    config = AppConfig()
    assert config.SuperName == "EnvName"



import pytest

from arfi_settings import ArFiSettings, EnvConfigDict


# @pytest.mark.current
@pytest.mark.env_file
def test_simple_env_file_data(
    simple_data_env_file,
    path_base_dir,
):
    class AppConfig(ArFiSettings):
        path_config_file: str = "default_path_config_file"

    config = AppConfig()
    assert config.path_config_file == ".env"


# @pytest.mark.current
@pytest.mark.env_file
def test_ordered_env_file_data(
    simple_data_env_file,
    simple_data_prod_env_file,
    path_base_dir,
):
    class AppConfig(ArFiSettings):
        path_config_file: str = "default_path_config_file"
        env_config = EnvConfigDict(
            env_file=[
                "prod.env",
                ".env",
            ],
        )

    config = AppConfig()
    assert config.path_config_file == ".env"

    class AppConfig(ArFiSettings):
        path_config_file: str = "default_path_config_file"
        env_config = EnvConfigDict(
            env_file=[
                ".env",
                "prod.env",
            ],
        )

    config = AppConfig()
    assert config.path_config_file == "prod.env"



import argparse
import sys
from pathlib import Path
from typing import Literal

import pytest

import arfi_settings
from arfi_settings import ArFiReader, ArFiSettings, EnvConfigDict, SettingsConfigDict
from arfi_settings.init_config import InitSettings


# @pytest.mark.current
@pytest.mark.pyproject
def test_init_settings_with_pyproject_toml(
    empty_pyproject_toml_file,
    cwd_to_tmp,
    monkeypatch,
):
    settings_file = cwd_to_tmp / "settings.py"
    settings_file.touch(exist_ok=True)

    class PatchedInitSettings(InitSettings):
        @staticmethod
        def _search_called_file():
            return settings_file.as_posix(), 0

    monkeypatch.setattr(arfi_settings.init_config, "InitSettings", PatchedInitSettings)
    init_settings = PatchedInitSettings()
    monkeypatch.setattr(arfi_settings.init_config, "init_settings", init_settings)
    monkeypatch.setattr(arfi_settings.main, "init_settings", init_settings)
    empty_pyproject_toml_file.write_text(
        """
        [tool.arfi_settings]
        env_file=''
        env_config_inherit_parent=false
        """
    )

    class AppSettings(ArFiSettings):
        pass

    class AppConfig(ArFiSettings):
        app: AppSettings
        env_config_inherit_parent = True

    assert init_settings.pyproject_toml_path is None
    assert init_settings.main_config_file is None
    assert init_settings.init_params.env_file == ".env"
    assert init_settings.init_params.env_config_inherit_parent is True

    _ = AppConfig()

    assert init_settings.pyproject_toml_path == empty_pyproject_toml_file
    assert init_settings.main_config_file == settings_file
    assert init_settings.init_params.env_file == ""
    assert init_settings.init_params.env_config_inherit_parent is False


# @pytest.mark.current
@pytest.mark.pyproject
def test_pyproject_toml_class_vars(
    cwd_to_tmp,
    monkeypatch,
):
    root_init_file = cwd_to_tmp / "__init__.py"
    root_init_file.touch(exist_ok=True)
    settings_dir = cwd_to_tmp / "settings"
    settings_dir.mkdir(exist_ok=True)
    settings_init_file = settings_dir / "__init__.py"
    settings_init_file.touch(exist_ok=True)
    settings_file = settings_dir / "settings.py"
    settings_file.touch(exist_ok=True)
    config_dir = cwd_to_tmp / "config"
    config_dir.mkdir(exist_ok=True)
    config_file = config_dir / "config.yml"
    config_file.touch(exist_ok=True)
    app_config_dir = config_dir / "app"
    app_config_dir.mkdir(exist_ok=True)
    app_config_file = app_config_dir / "config.yml"
    app_config_file.touch(exist_ok=True)
    app_config_file.write_text(
        """
        mode: 'prod'
        path_config_file: 'config/app/config.yml'
        """
    )
    app_mode_prod_config_file = app_config_dir / "prod.json"
    app_mode_prod_config_file.touch(exist_ok=True)
    app_mode_prod_config_file.write_text(
        """
        {
            "path_config_file": "config/app/prod.json",
            "database_dialect": "postgresql"
        }
        """
    )
    app_mode_dev_config_file = app_config_dir / "dev.json"
    app_mode_dev_config_file.touch(exist_ok=True)
    app_mode_dev_config_file.write_text(
        """
        {
            "path_config_file": "config/app/dev.json",
            "database_dialect": "mysql"
        }
        """
    )
    app_mode_test_config_file = app_config_dir / "test.yml"
    app_mode_test_config_file.touch(exist_ok=True)
    app_mode_test_config_file.write_text(
        """
        database_dialect: "sqlite"
        """
    )

    class PatchedInitSettings(InitSettings):
        @staticmethod
        def _search_called_file():
            return settings_file.as_posix(), 0

    monkeypatch.setattr(arfi_settings.init_config, "InitSettings", PatchedInitSettings)
    init_settings = PatchedInitSettings()
    monkeypatch.setattr(arfi_settings.init_config, "init_settings", init_settings)
    monkeypatch.setattr(arfi_settings.main, "init_settings", init_settings)
    monkeypatch.setattr(arfi_settings.readers.ArFiReader, "default_cli_reader", None)

    pyproject_toml_file = cwd_to_tmp / "pyproject.toml"
    pyproject_toml_file.touch(exist_ok=True)
    pyproject_toml_file.write_text(
        """
        [tool.arfi_settings]
        env_file=''
        env_config_inherit_parent=false
        env_file_encoding="cp1251"
        ignore_missing=true
        conf_ignore_missing=false
        conf_exclude_inherit_parent=[
            "conf_ignore_missing",
            "conf_ext"
        ]
        secrets_dir="secrets/"
        """
    )
    secrets_dir = cwd_to_tmp / "secrets"
    secrets_dir.mkdir(exist_ok=True)

    class AppSettings(ArFiSettings):
        path_config_file: str
        database_dialect: Literal["sqlite", "postgresql", "mysql"] = "mysql"
        model_config = SettingsConfigDict(
            env_file_encoding="utf-8",
            conf_ext=["yml", "json"],
            env_file=".env.prod",
            cli=True,
        )

    class AppConfig(ArFiSettings):
        app: AppSettings
        env_config_inherit_parent = True
        env_config = EnvConfigDict(
            env_case_sensitive=True,
        )
        model_config = SettingsConfigDict(
            ignore_missing=False,
            conf_ignore_missing=True,
        )

    assert init_settings.pyproject_toml_path is None
    assert init_settings.main_config_file is None
    assert init_settings.init_params.env_file == ".env"
    assert init_settings.init_params.env_config_inherit_parent is True

    config = AppConfig()

    assert init_settings.pyproject_toml_path == pyproject_toml_file
    assert init_settings.main_config_file == settings_file
    assert init_settings.init_params.env_file == ""
    assert init_settings.init_params.env_config_inherit_parent is False

    assert config.pyproject_toml_path == pyproject_toml_file
    assert config.app.pyproject_toml_path == pyproject_toml_file
    assert config.env_config_inherit_parent is True
    assert config.app.env_config_inherit_parent is False
    assert config.settings_config.env_file == ""
    assert config.app.settings_config.env_file == ".env.prod"
    assert config.settings_config.env_file_encoding == "cp1251"
    assert config.app.settings_config.env_file_encoding == "utf-8"

    assert config.settings_config.conf_ignore_missing is True
    assert config.settings_config.env_ignore_missing is False
    assert config.app.settings_config.conf_ignore_missing is False
    assert config.app.settings_config.env_ignore_missing is True

    assert config.settings_config.conf_ext == ["toml", "yaml", "yml", "json"]
    assert config.app.settings_config.conf_ext == ["yml", "json"]

    assert config.app.path_config_file == "config/app/prod.json"
    assert config.app.database_dialect == "postgresql"

    secrets_file = secrets_dir / "path_config_FILE"
    secrets_file.touch(exist_ok=True)
    secrets_file.write_text("secrets/path_config_file")

    config = AppConfig()
    assert config.settings_config.secrets_dir == Path("secrets")
    assert config.settings_config.secrets_dir.resolve() == secrets_dir
    assert config.app.path_config_file == "secrets/path_config_file"

    env_file = cwd_to_tmp / ".env.prod"
    env_file.touch(exist_ok=True)
    env_file.write_text("PATH_CONFIG_FILE=ENV_FILE_PATH_CONFIG_FILE")
    config = AppConfig()
    assert config.app.path_config_file == "ENV_FILE_PATH_CONFIG_FILE"

    monkeypatch.setenv("PATH_CONFIG_FILE", "env_path_config_file")
    config = AppConfig()
    assert config.app.path_config_file == "env_path_config_file"

    def parse_args():
        parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        parser.add_argument(
            "--debug",
            type=bool,
            const=True,
            nargs="?",
            help="Debug application",
        )
        parser.add_argument(
            "--mode",
            type=str,
            help="Application mode",
        )
        cli_options = parser.parse_args()
        data = dict(cli_options._get_kwargs())
        if not cli_options.mode:
            data.pop("mode")
        if cli_options.debug and not cli_options.mode:
            data["MODE"] = "dev"
        return data

    ArFiReader.setup_cli_reader(parse_args)
    monkeypatch.setattr(sys, "argv", [sys.argv[0]])
    config = AppConfig()
    assert config.MODE is None
    assert config.app.MODE == "prod"
    assert config.app.path_config_file == "env_path_config_file"
    assert config.app.database_dialect == "postgresql"

    monkeypatch.setattr(sys, "argv", [sys.argv[0], "--debug"])
    config = AppConfig()
    assert config.MODE is None
    assert config.app.MODE == "dev"
    assert config.app.path_config_file == "env_path_config_file"
    assert config.app.database_dialect == "mysql"

    monkeypatch.setattr(sys, "argv", [sys.argv[0], "--debug", "--mode", "test"])
    config = AppConfig()
    assert config.MODE is None
    assert config.app.MODE == "test"
    assert config.app.path_config_file == "env_path_config_file"
    assert config.app.database_dialect == "sqlite"



from enum import StrEnum


class FileConfigEntity(StrEnum):
    """Entity type of a file config settings."""

    DEFAULT = "config"
    FILE = "conf_file"
    DIR = "conf_dir"
    EXT = "conf_ext"
    FILE_ENCODING = "conf_file_encoding"
    CASE_SENSITIVE = "conf_case_sensitive"
    IGNORE_MISSING = "conf_ignore_missing"
    CUSTOM_EXT_HANDLER = "conf_custom_ext_handler"
    CONF_INCLUDE_INHERIT_PARENT = "conf_include_inherit_parent"
    CONF_EXCLUDE_INHERIT_PARENT = "conf_exclude_inherit_parent"


class EnvConfigEntity(StrEnum):
    """Entity type of an env config settings."""

    FILE = "env_file"
    PREFIX = "env_prefix"
    PREFIX_AS_MODE_DIR = "env_prefix_as_mode_dir"
    PREFIX_AS_NESTED_MODE_DIR = "env_prefix_as_nested_mode_dir"
    PREFIX_AS_SOURCE_MODE_DIR = "env_prefix_as_source_mode_dir"
    FILE_ENCODING = "env_file_encoding"
    CASE_SENSITIVE = "env_case_sensitive"
    NESTED_DELIMITER = "env_nested_delimiter"
    DEFAULT_ENV_FILE = ".env"
    IGNORE_MISSING = "env_ignore_missing"
    ENV_INCLUDE_INHERIT_PARENT = "env_include_inherit_parent"
    ENV_EXCLUDE_INHERIT_PARENT = "env_exclude_inherit_parent"


class ModelConfigEntity(StrEnum):
    """Entity type of a model config settings."""

    CASE_SENSITIVE = "case_sensitive"
    IGNORE_MISSING = "ignore_missing"
    ENCODING = "encoding"
    CLI = "cli"
    SECRETS_DIR = "secrets_dir"
    INCLUDE_INHERIT_PARENT = "include_inherit_parent"
    EXCLUDE_INHERIT_PARENT = "exclude_inherit_parent"



from pathlib import Path

import pytest

from arfi_settings import ArFiSettings
from arfi_settings.types import DEFAULT_PATH_SENTINEL


# @pytest.mark.current
@pytest.mark.mode_dir
def test_nested_mode_dir(cwd_to_tmp):
    class Database(ArFiSettings):
        mode_dir = "db"

    class Postgres(Database):
        mode_dir = "postgres"

    class AppConfig(ArFiSettings):
        # read_config = True
        db: Postgres = Postgres()

    config = AppConfig()
    assert config.db.mode_dir == "db/postgres"


# @pytest.mark.current
@pytest.mark.mode_dir
def test_mode_dir_inherit_nested(cwd_to_tmp):
    class Database(ArFiSettings):
        mode_dir = "db"

    class Postgres(Database):
        mode_dir = "postgres"
        mode_dir_inherit_nested = False

    class AppConfig(ArFiSettings):
        read_config = True
        db: Postgres = Postgres()

    config = AppConfig()
    assert config.db.mode_dir == "postgres"


# @pytest.mark.current
@pytest.mark.mode_dir
def test_mode_dir_inherit_nested_false(cwd_to_tmp):
    class Database(ArFiSettings):
        mode_dir = "db"

    class Postgres(Database):
        mode_dir_inherit_nested = False

    class AppConfig(ArFiSettings):
        read_config = True
        db: Database = Postgres()

    config = AppConfig()
    assert config.db.mode_dir == "db"


# @pytest.mark.current
@pytest.mark.mode_dir
def test_mode_dir_by_attr(cwd_to_tmp):
    class Proxy(ArFiSettings):
        pass

    class AppSettings(ArFiSettings):
        proxy: Proxy = Proxy()

    class AppConfig(ArFiSettings):
        read_config = True
        app: AppSettings = AppSettings()

    config = AppConfig()
    assert config.app.mode_dir == "app"
    assert config.app.proxy.mode_dir == "proxy"
    assert config.app.proxy.mode_dir_path == Path("app/proxy")
    app_settings = AppSettings()
    assert app_settings.mode_dir == DEFAULT_PATH_SENTINEL


# @pytest.mark.current
@pytest.mark.mode_dir
def test_mode_dir_by_attr_inherit_parent_false(cwd_to_tmp):
    class Proxy(ArFiSettings):
        mode_dir_inherit_parent = False

    class AppSettings(ArFiSettings):
        proxy: Proxy = Proxy()

    class AppConfig(ArFiSettings):
        read_config = True
        mode_dir = "dev"
        app: AppSettings = AppSettings()

    config = AppConfig()
    assert config.app.mode_dir == "app"
    assert config.app.proxy.mode_dir == "proxy"
    assert config.app.mode_dir_path == Path("dev/app")
    assert config.app.proxy.mode_dir_path == Path("proxy")
    app_settings = AppSettings()
    assert app_settings.mode_dir == DEFAULT_PATH_SENTINEL


# @pytest.mark.current
@pytest.mark.mode_dir
def test_init_mode_dir(cwd_to_tmp):
    class Database(ArFiSettings):
        mode_dir = "db"

    class Postgres(Database):
        mode_dir = "postgres"

    class AppConfig(ArFiSettings):
        read_config = True
        db: Postgres
        pg: Postgres = Postgres(_mode_dir="pg")
        prod_postgres: Postgres = Postgres(_mode_dir="prod_postgres")
        test_db: Postgres = Postgres()
        mdb: Postgres = Postgres(_mode_dir="mdb", _mode_dir_inherit_nested=False)

    config = AppConfig()
    assert config.db.mode_dir == "db/postgres"
    assert config.pg.mode_dir == "db/pg"
    assert config.prod_postgres.mode_dir == "db/prod_postgres"
    assert config.test_db.mode_dir == "db/postgres"
    assert config.mdb.mode_dir == "mdb"

    pg_conf = Postgres(_mode_dir="pg", MODE="pg")
    assert pg_conf.read_config
    assert pg_conf.MODE == "pg"
    assert pg_conf.mode_dir == "db/pg"


# @pytest.mark.current
@pytest.mark.mode_dir
def test_nested_with_init_mode_dir(cwd_to_tmp):
    class Database(ArFiSettings):
        mode_dir = "db"

    class Postgres(Database):
        mode_dir = "postgres"

    class AppConfig(ArFiSettings):
        read_config = True
        db: Postgres = Postgres()

    config = AppConfig()
    assert config.db.mode_dir == "db/postgres"

    class QuerySettings(ArFiSettings):
        db: Postgres = Postgres(_mode_dir="query")

    q = QuerySettings()
    assert q.db.mode_dir == "db/query"


# @pytest.mark.current
@pytest.mark.mode_dir
def test_parent_mode_dir(cwd_to_tmp):
    class Proxy(ArFiSettings):
        pass

    class AppSettings(ArFiSettings):
        proxy: Proxy

    class AppConfig(ArFiSettings):
        # read_config = True
        app: AppSettings

    config = AppConfig()
    assert config.app.mode_dir == "app"
    assert config.app.proxy.parent_mode_dir == "app"
    assert config.app.proxy.mode_dir_path == Path("app/proxy")

    # print()
    # print("=*" * 50)
    # print(config.model_dump_json(indent=4))
    # print(f"{config.app.conf_path = }")
    # print(f"{config.app.proxy.conf_path = }")


# @pytest.mark.current
@pytest.mark.mode_dir
def test_mode_dir_inherit_parent(cwd_to_tmp):
    class Proxy(ArFiSettings):
        mode_dir_inherit_parent = False

    class AppSettings(ArFiSettings):
        proxy: Proxy = Proxy()

    class AppConfig(ArFiSettings):
        read_config = True
        app: AppSettings = AppSettings()

    config = AppConfig()
    assert config.app.mode_dir == "app"
    assert config.app.proxy.parent_mode_dir == "app"
    assert config.app.proxy.mode_dir_path == Path("proxy")


# @pytest.mark.current
@pytest.mark.mode_dir
def test_empty_mode_dir(cwd_to_tmp):
    class Proxy(ArFiSettings):
        pass

    class AppSettings(ArFiSettings):
        mode_dir = ""
        proxy: Proxy = Proxy()

    class AppConfig(ArFiSettings):
        read_config = True
        app: AppSettings = AppSettings()

    config = AppConfig()
    assert config.app.proxy.parent_mode_dir == DEFAULT_PATH_SENTINEL
    assert config.app.proxy.mode_dir_path == Path("proxy")


# @pytest.mark.current
@pytest.mark.mode_dir
def test_none_mode_dir(cwd_to_tmp):
    class Proxy(ArFiSettings):
        pass

    class AppSettings(ArFiSettings):
        mode_dir = None
        proxy: Proxy = Proxy()

    class AppConfig(ArFiSettings):
        read_config = True
        app: AppSettings = AppSettings()

    config = AppConfig()
    assert config.app.proxy.parent_mode_dir == DEFAULT_PATH_SENTINEL
    assert config.app.proxy.mode_dir_path == Path("proxy")


@pytest.mark.mode_dir
def test_empty_mode_dir_with_parent_mode_dir(cwd_to_tmp):
    class Proxy(ArFiSettings):
        pass

    class AppSettings(ArFiSettings):
        mode_dir = ""
        proxy: Proxy = Proxy()

    class AppConfig(ArFiSettings):
        read_config = True
        mode_dir = "dev"
        app: AppSettings = AppSettings()

    config = AppConfig()
    assert config.app.parent_mode_dir == "dev"
    assert config.app.mode_dir_path == Path("dev")
    assert config.app.proxy.parent_mode_dir == "dev"
    assert config.app.proxy.mode_dir_path == Path("dev/proxy")


# @pytest.mark.current
@pytest.mark.mode_dir
def test_none_mode_dir_with_parent_mode_dir(cwd_to_tmp):
    class Proxy(ArFiSettings):
        pass

    class AppSettings(ArFiSettings):
        mode_dir = None
        proxy: Proxy = Proxy()

    class AppConfig(ArFiSettings):
        read_config = True
        mode_dir = "dev"
        app: AppSettings = AppSettings()

    config = AppConfig()
    assert config.app.parent_mode_dir == "dev"
    assert config.app.mode_dir_path == Path("dev")
    assert config.app.proxy.parent_mode_dir == "dev"
    assert config.app.proxy.mode_dir_path == Path("dev/proxy")


# @pytest.mark.current
@pytest.mark.mode_dir
def test_read_config(cwd_to_tmp):
    class Proxy(ArFiSettings):
        pass

    class AppSettings(ArFiSettings):
        proxy: Proxy = Proxy()

    class AppConfig(ArFiSettings):
        read_config = True
        app: AppSettings = AppSettings()

    config = AppConfig()
    assert config.read_config


@pytest.mark.mode_dir
@pytest.mark.xfail
@pytest.mark.parametrize("value", [None, 0, 1, "", "some_string", "False"])
def test_read_config_not_bool(cwd_to_tmp, value):
    class Proxy(ArFiSettings):
        pass

    class AppSettings(ArFiSettings):
        proxy: Proxy = Proxy()

    class AppConfig(ArFiSettings):
        read_config = value
        app: AppSettings = AppSettings()

    config = AppConfig()
    assert config.read_config


# @pytest.mark.current
@pytest.mark.mode_dir
@pytest.mark.parametrize(
    "values, app_mode, proxy_mode",
    [
        (
            {"app": {"MODE": "my_app_mode", "proxy": {"MODE": "my_proxy_mode"}}},
            "my_app_mode",
            "my_proxy_mode",
        ),
        (
            {"app": {"proxy": {}}},
            None,
            None,
        ),
    ],
)
def test_mode(cwd_to_tmp, values, app_mode, proxy_mode, path_base_dir):
    class Proxy(ArFiSettings):
        pass

    class AppSettings(ArFiSettings):
        proxy: Proxy = Proxy()

    class AppConfig(ArFiSettings):
        read_config = True
        app: AppSettings = AppSettings()

    config = AppConfig(**values)
    assert config.MODE is None
    assert config.app.MODE == app_mode
    assert config.app.proxy.MODE == proxy_mode
    assert config.app.mode_dir == "app"
    assert config.app.proxy.mode_dir == "proxy"



import inspect
from pathlib import Path
from typing import Any, ClassVar, Literal

from pydantic import AliasChoices, BaseModel, Field

from .arfi_debug import debug
from .constants import (
    PYPROJECT_TOML_MAX_DEPTH,
)
from .descriptors import (
    BaseDirDescriptor,
    EnvConfigInheritParentDescriptor,
    FileConfigInheritParentDescriptor,
    HandlerDescriptor,
    HandlerInheritParentDescriptor,
    ModeDirDescriptor,
    ModeDirInheritNestedDescriptor,
    ModeDirInheritParentDescriptor,
    OrderedSettingsDescriptor,
    OrderedSettingsInheritParentDescriptor,
    ReadConfigDeskriptor,
    ReadConfigForceDeskriptor,
    ReadPyProjectTomlDeskriptor,
)
from .handlers import ArFiHandler
from .init_config import init_settings
from .schemes import (
    CONF_INCLUDE_EXLUDE_PARAMS,
    ENV_INCLUDE_EXLUDE_PARAMS,
    INCLUDE_EXLUDE_PARAMS,
    SettingsConfigSchema,
)
from .storage import config_storage
from .types import (
    DEFAULT_PATH_SENTINEL,
    LIST_STR_SENTINEL,
    SENTINEL,
    STR_SENTINEL,
    EnvConfigDict,
    FileConfigDict,
    MultiPathType,
    PathType,
    SettingsConfigDict,
)
from .utils import is_descriptor


class ArFiSettings(BaseModel):
    """Advanced pydantic settings."""

    MODE: str | None = Field(None, validation_alias=AliasChoices("MODE", "mode"))
    BASE_DIR: ClassVar[PathType | None] = None
    read_config: ClassVar[bool] = True
    read_config_force: ClassVar[bool] = None
    read_pyproject_toml: ClassVar[bool] = True
    mode_dir: ClassVar[PathType | None] = DEFAULT_PATH_SENTINEL
    mode_dir_inherit_nested: ClassVar[bool] = SENTINEL
    mode_dir_inherit_parent: ClassVar[bool] = SENTINEL
    file_config_inherit_parent: ClassVar[bool] = SENTINEL
    env_config_inherit_parent: ClassVar[bool] = SENTINEL
    file_config: ClassVar[FileConfigDict] = FileConfigDict()
    env_config: ClassVar[EnvConfigDict] = EnvConfigDict()
    handler_class: ClassVar[ArFiHandler] = ArFiHandler
    handler: ClassVar[str] = SENTINEL
    handler_inherit_parent: ClassVar[bool] = SENTINEL
    ordered_settings: ClassVar[list[str]] = SENTINEL
    ordered_settings_inherit_parent: ClassVar[bool] = SENTINEL

    model_config: ClassVar[SettingsConfigDict] = SettingsConfigDict(
        extra="forbid",
        arbitrary_types_allowed=True,
        validate_default=True,
    )

    __slots__ = (
        "_read_config",
        "_read_config_force",
        "_read_pyproject_toml",
        "_init_file_config",
        "_init_env_config",
        "_settings_init_params",
        "_mode_dir",
        "_mode_dir_inherit_nested",
        "_mode_dir_inherit_parent",
        "_file_config_inherit_parent",
        "_env_config_inherit_parent",
        "_handler_inherit_parent",
        "_ordered_settings",
        "_ordered_settings_inherit_parent",
        "_values_by_default",
        "_handler_tree",
        "_handler_parent_mode_dir",
        "_mode_dir_attr",
        "_parent_mode_dir",
        "_parent_file_config",
        "_parent_env_config",
        "_handler",
        "_conf_path",
        "_env_path",
        "_computed_file_config",
        "_computed_env_config",
        "_settings_config",
        "_computed_mode_dir",
        "_mode_dir_path",
        "_nested_mode_dir",
        "_source_mode_dir",
        "_inherited_params",
        "_modified_class_vars",
        "_pyproject_toml_path",
        "_search_base_dir",
        "_base_dir",
        "_root_dir",
        "_arfi_debug",
    )

    __instances: ClassVar[dict] = dict()

    @property
    def settings_config(self) -> SettingsConfigSchema:
        if not hasattr(self, "_settings_config"):
            return None
        return self._settings_config

    @property
    def inherited_params(self) -> list[str]:
        if not hasattr(self, "_inherited_params"):
            return []
        return self._inherited_params

    # @computed_field
    @property
    def conf_path(self) -> list[Path]:
        """Full config path from file and dir settings."""
        if not hasattr(self, "_conf_path"):
            return []
        return self._conf_path

    # @computed_field
    @property
    def env_path(self) -> list[Path]:
        """Full env path from file and dir settings."""
        if not hasattr(self, "_env_path"):
            return []
        return self._env_path

    @property
    def mode_dir_path(self) -> PathType:
        """Generate path to mode directory."""
        if not hasattr(self, "_mode_dir_path"):
            return DEFAULT_PATH_SENTINEL
        return self._mode_dir_path

    @property
    def computed_mode_dir(self) -> str:
        if not hasattr(self, "_computed_mode_dir"):
            return STR_SENTINEL
        return self._computed_mode_dir

    @property
    def source_mode_dir(self) -> str:
        if not hasattr(self, "_source_mode_dir"):
            return STR_SENTINEL
        return self._source_mode_dir

    @property
    def nested_mode_dir(self) -> PathType | None:
        if not hasattr(self, "_nested_mode_dir"):
            return DEFAULT_PATH_SENTINEL
        return self._nested_mode_dir

    @property
    def parent_mode_dir(self) -> PathType | None:
        if not hasattr(self, "_parent_mode_dir"):
            return DEFAULT_PATH_SENTINEL
        return self._parent_mode_dir

    @property
    def computed_file_config(self) -> FileConfigDict:
        """Compute file_config."""
        if not hasattr(self, "_computed_file_config"):
            return FileConfigDict()
        return self._computed_file_config

    @property
    def computed_env_config(self) -> EnvConfigDict:
        """Compute env_config."""
        if not hasattr(self, "_computed_env_config"):
            return EnvConfigDict()
        return self._computed_env_config

    @property
    def pyproject_toml_path(self) -> PathType | None:
        """The path to pyproject.toml, defined during instance initialization."""
        if not hasattr(self, "_pyproject_toml_path"):
            return None
        return self._pyproject_toml_path

    @property
    def root_dir(self) -> PathType | None:
        return self._root_dir

    def __set_name__(self, owner, name):
        self._mode_dir_attr = name

    def __setattr__(self, attr, value):
        if attr in self.__slots__:
            object.__setattr__(self, attr, value)
        else:
            super().__setattr__(attr, value)

    def __init_subclass__(cls):
        super().__init_subclass__()
        cls._mode_dir_attr = None
        cls._modified_class_vars: set[str] = set()
        cls._search_base_dir = True
        cls._root_dir = None
        cls._base_dir = None
        cls._arfi_debug = False
        cls._setup_subclasses_descriptors()

    @classmethod
    def _setup_subclasses_descriptors(cls):
        """Validate and set class variables as descriptors"""

        # setup settings from users class, from pyproject.toml or by default
        for cls_var in init_settings.init_params.model_fields:
            if cls_var not in cls.__class_vars__:
                continue
            privat_name = f"_{cls_var}"
            value = getattr(cls, cls_var)
            if value is SENTINEL:
                value = getattr(init_settings.init_params, cls_var)
                setattr(cls, privat_name, value)
            else:
                cls._modified_class_vars.add(cls_var)
                setattr(cls, privat_name, value)
        cls._base_dir = cls.BASE_DIR
        if cls._base_dir is not None:
            cls._base_dir = Path(cls._base_dir).expanduser().resolve()
            cls._search_base_dir = False
        cls.BASE_DIR = BaseDirDescriptor(cls._base_dir)
        cls._read_config = cls.read_config
        cls.read_config = ReadConfigDeskriptor(cls._read_config)
        cls._read_config_force = cls.read_config_force
        cls.read_config_force = ReadConfigForceDeskriptor(cls._read_config_force)
        cls._read_pyproject_toml = cls.read_pyproject_toml
        cls.read_pyproject_toml = ReadPyProjectTomlDeskriptor(cls._read_pyproject_toml)
        if is_descriptor(cls, "mode_dir"):
            # The classic inheritance of the `mode_dir` parameter is disabled,
            # since it participates in reverse inheritance.
            cls.mode_dir = DEFAULT_PATH_SENTINEL
        cls._mode_dir = cls.mode_dir
        cls.mode_dir = ModeDirDescriptor(cls._mode_dir)
        cls.mode_dir_inherit_parent = ModeDirInheritParentDescriptor(cls._mode_dir_inherit_parent)
        cls.mode_dir_inherit_nested = ModeDirInheritNestedDescriptor(cls._mode_dir_inherit_nested)
        cls.file_config_inherit_parent = FileConfigInheritParentDescriptor(cls._file_config_inherit_parent)
        cls.env_config_inherit_parent = EnvConfigInheritParentDescriptor(cls._env_config_inherit_parent)
        cls.handler = HandlerDescriptor(cls._handler)
        cls.handler_inherit_parent = HandlerInheritParentDescriptor(cls._handler_inherit_parent)
        cls.ordered_settings = OrderedSettingsDescriptor(cls._ordered_settings)
        cls.ordered_settings_inherit_parent = OrderedSettingsInheritParentDescriptor(
            cls._ordered_settings_inherit_parent
        )

    def __new__(cls, _instance_id: int = None, **kwargs):
        """Create new instance or return existing."""

        curren_frame = inspect.currentframe()
        parrent_frame = curren_frame.f_back
        frame_scope = parrent_frame.f_locals
        co_name = parrent_frame.f_code.co_name

        given_instance = cls.__instances.get(_instance_id)
        if given_instance and not issubclass(type(given_instance), cls):
            given_instance = None

        if given_instance:
            instance = given_instance
            instance._read_config = True
        else:
            instance = super().__new__(cls)
            if "__qualname__" in frame_scope:
                instance._read_config = False
                instance._read_pyproject_toml = False
            if co_name not in ("__init__", "__deepcopy__"):
                cls.__instances[id(instance)] = instance

            if co_name == "__deepcopy__":
                instance._read_config = False
                instance._read_config_force = False
                instance._read_pyproject_toml = False

        return instance

    def _convert_debug_path(self, list_path: list[Path], mode: Literal["conf", "env"]) -> list[str]:
        result = []
        base_dir = self.BASE_DIR
        root_dir = self.root_dir
        if self.BASE_DIR is not None:
            base_dir = Path(self.BASE_DIR).resolve()
        if self.root_dir is not None:
            root_dir = Path(self.root_dir).resolve()
        for path in list_path:
            if mode == "conf":
                if base_dir is not None:
                    path = base_dir / path
            elif mode == "env":
                if root_dir is not None:
                    if Path(root_dir / path).is_file():
                        path = root_dir / path
                    else:
                        if base_dir is not None:
                            path = base_dir / path
                else:
                    if base_dir is not None:
                        path = base_dir / path
            result.append(path.resolve().as_posix())
        return result

    @staticmethod
    def _clear_value_from_handler_params(value: dict[str, Any]) -> dict[str, Any]:
        """Remove handler auxiliary params from value."""

        extra_fields = [
            "_instance_id",
            "_init_value",
            "_handler_value",
            "_handler_tree",
            "_handler_parent_mode_dir",
            "_handler_mode_dir_attr",
            "_handler_ordered_settings",
            "_handler_read_pyproject_toml_force",
            "_handler_search_base_dir",
            "_handler_main_handler",
        ]
        for field in extra_fields:
            if field in value:
                value.pop(field)
        return value

    def __init__(
        self,
        _read_config: bool = None,
        _read_config_force: bool = None,
        _read_pyproject_toml: bool | None = None,
        _pyproject_toml_depth: int | None = None,
        _pyproject_toml_max_depth: int = PYPROJECT_TOML_MAX_DEPTH,
        _mode_dir: PathType | None = DEFAULT_PATH_SENTINEL,
        _mode_dir_inherit_nested: bool = None,
        _mode_dir_inherit_parent: bool = None,
        _file_config_inherit_parent: bool = None,
        _env_config_inherit_parent: bool = None,
        _conf_file: MultiPathType | None = DEFAULT_PATH_SENTINEL,
        _conf_dir: MultiPathType | None = DEFAULT_PATH_SENTINEL,
        _conf_ext: str | list[str, ...] = [],
        _conf_file_encoding: str | None = STR_SENTINEL,
        _conf_case_sensitive: bool = None,
        _conf_ignore_missing: bool = None,
        _conf_custom_ext_handler: str | dict[str, str] | None = STR_SENTINEL,
        _conf_include_inherit_parent: list[Literal[*CONF_INCLUDE_EXLUDE_PARAMS], ...] = None,
        _conf_exclude_inherit_parent: list[Literal[*CONF_INCLUDE_EXLUDE_PARAMS], ...] = None,
        _env_file: MultiPathType | None = DEFAULT_PATH_SENTINEL,
        _env_prefix: str = None,
        _env_prefix_as_mode_dir: bool = None,
        _env_prefix_as_nested_mode_dir: bool = None,
        _env_prefix_as_source_mode_dir: bool = None,
        _env_file_encoding: str | None = STR_SENTINEL,
        _env_case_sensitive: bool = None,
        _env_nested_delimiter: str = None,
        _env_ignore_missing: bool = None,
        _env_include_inherit_parent: list[Literal[*ENV_INCLUDE_EXLUDE_PARAMS], ...] = None,
        _env_exclude_inherit_parent: list[Literal[*ENV_INCLUDE_EXLUDE_PARAMS], ...] = None,
        _case_sensitive: bool = None,
        _ignore_missing: bool = None,
        _encoding: str | None = STR_SENTINEL,
        _secrets_dir: PathType | None = DEFAULT_PATH_SENTINEL,
        _include_inherit_parent: list[Literal[*INCLUDE_EXLUDE_PARAMS], ...] = None,
        _exclude_inherit_parent: list[Literal[*INCLUDE_EXLUDE_PARAMS], ...] = None,
        _handler: str = STR_SENTINEL,
        _handler_inherit_parent: bool = None,
        _ordered_settings: list[str, ...] = LIST_STR_SENTINEL,
        _ordered_settings_inherit_parent: bool = None,
        _cli: bool = None,
        _arfi_debug: bool = False,
        **values: Any,
    ):
        _kwargs: dict = locals()
        _kwargs.pop("self")
        config = config_storage.get_config(self)
        values = config.load(instance=self, **_kwargs)
        config.extract(self)

        debug(self, "__init__", arfi_debug=_arfi_debug, mode="before", values=values)

        if self.read_config:
            handler = self.handler_class(
                settings_class=self,
                init_kwargs=values,
                handler=self.handler,
            )
            values = handler()
        values = self._clear_value_from_handler_params(values)

        debug(self, "__init__", values=values, mode="after")

        super().__init__(**values)
        config.extract(self)



from typing import Literal

import pytest
from pydantic import AliasChoices, BaseModel, Field

from arfi_settings import ArFiSettings


# @pytest.mark.current
@pytest.mark.field_default_value
def test_default_value(cwd_to_tmp, path_base_dir):
    class AppSettings(ArFiSettings):
        pass

    class AppConfig(ArFiSettings):
        app: AppSettings = AppSettings()

    config = AppConfig()
    assert config.MODE is None
    assert config.app.MODE is None

    class AppConfig(ArFiSettings):
        app: AppSettings = AppSettings(mode="prod")

    config = AppConfig()
    assert config.MODE is None
    assert config.app.MODE == "prod"

    class AppSettings(ArFiSettings):
        MODE: str | None = Field("prod", validation_alias=AliasChoices("mode", "MODE"))

    class AppConfig(ArFiSettings):
        app: AppSettings = AppSettings()

    config = AppConfig()
    assert config.MODE is None
    assert config.app.MODE == "prod"

    app = AppSettings(mode="dev")
    config = AppConfig(app=app)
    assert config.app.MODE == "dev"

    config = AppConfig(app=AppSettings(mode="test"))
    assert config.app.MODE == "test"

    class PydanticConfig(BaseModel):
        app: AppSettings = AppSettings(mode="pydantic")

    config = PydanticConfig()
    assert config.app.MODE == "pydantic"


# @pytest.mark.current
@pytest.mark.field_default_value
def test_default_value_with_discriminator(cwd_to_tmp, path_base_dir):
    class SQLite(BaseModel):
        db_type: Literal["sqlite"] = "sqlite"

    class Postgre(ArFiSettings):
        db_type: Literal["postgre"] = "postgre"

    class MySQL(ArFiSettings):
        db_type: Literal["mysql"] = "mysql"

    class AppConfig(ArFiSettings):
        db: SQLite | Postgre = Field(  # SQLite is not settings field type
            default=Postgre(mode="prod"),
            discriminator="db_type",
        )

    config = AppConfig()
    assert config.db.MODE == "prod"
    assert config.db.db_type == "postgre"

    data = {"db": {"mode": "asd", "db_type": "postgre"}}
    config = AppConfig(**data)
    assert config.db.MODE == "asd"
    assert config.db.db_type == "postgre"

    data = {"db": {"db_type": "sqlite"}}
    config = AppConfig(**data)
    assert config.db.db_type == "sqlite"

    # the default values are not readable
    data = {"db": {"db_type": "postgre"}}
    config = AppConfig(**data)
    assert config.db.MODE is None  # because db: SQLite is not settings field type
    assert config.db.db_type == "postgre"

    class AppConfig(ArFiSettings):
        db: MySQL | Postgre = Field(  # all types are settings field types
            default=Postgre(mode="prod"),
            discriminator="db_type",
        )

    # the default values are read
    data = {"db": {"db_type": "postgre"}}
    config = AppConfig(**data)
    assert config.db.MODE == "prod"  # because db: MySQL is settings field type
    assert config.db.db_type == "postgre"
